// main.c: Defines the entry point for the console application.
//

#define _CRT_SECURE_NO_WARNINGS

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <time.h>

#include "../pinpad/pinpad.h"
#include "../pinpad/bertlv.h"

#include "rs232.h"
#include "des.h"
#include "tags.h"
#include "backend.h"

// Print verbose debug information
#define VERBOSE 0

// Macro that verify result and print debug inforation ogn error. 
#define CHECK_RESULT(X)												\
	do { unsigned int err = (X);									\
		if (err != PPSDK_ENOERR) {									\
			fprintf(stderr, "" #X " line: %d failed: %s\n",			\
				 __LINE__, ppsdk_getErrorText(err));				\
			return FALSE;											\
		}															\
	} while (0);

	
///////////////////////////////////////////////////////////////////////////////
// Local definitions.
///////////////////////////////////////////////////////////////////////////////
#define DEFAULT_SERIAL_PORT             "COM9:"
#define MAX_TRANSACTION_TIME            60000
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Local variables.
///////////////////////////////////////////////////////////////////////////////
static BOOL isTransactionActive = FALSE; // Transaction state
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Helper methods
///////////////////////////////////////////////////////////////////////////////
// Return last error text representation
static char* geterror()
{
    static char* msg = NULL;

	if (msg) LocalFree(msg);

    // Ask Windows to prepare a standard message for a GetLastError() code:
    FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, 
		NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPSTR)&msg, 0, NULL);
    
	// Return the message
	if (!msg) {
        return("Unknown error");
	}    
    
	return(msg);
}
//-----------------------------------------------------------------------------
// Print buffer content
static void printb(const unsigned char *buf, int len)
{
	// Heximal nibs
	static const char hex[] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };		
	char tmp[4];
	int n;

	for (n = 0; n < len; n++)
	{
		tmp[0] = hex[(buf[n]) >> 4];
		tmp[1] = hex[(buf[n]) & 15];
		tmp[2] = ' ';
		tmp[3] = '\0';

		if (!(n % 16)) {
			printf("\n");
		}
		printf(tmp);		
	}	
}

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Pinpad callback definitions
///////////////////////////////////////////////////////////////////////////////
//-----------------------------------------------------------------------------
static unsigned int read(void *param, unsigned char out[], unsigned int outMaxSize,
						 unsigned int *nbr, unsigned int timeout)
{
	DWORD complete = GetTickCount() + timeout;	

	do {
		BOOL res = rs232_read((HANDLE)param, out, outMaxSize, nbr);

		if (res) {
			if ((*nbr) > 0) {
				if (VERBOSE) { 
					printf("<< read(%d): ", *nbr); printb(out, *nbr); printf("\n");
				}				
			} else {
				if ((long)(complete) - (long)(GetTickCount()) < 0) {				
					return PPSDK_ETIMEOUT;
				}
			}
		} else {
			fprintf(stderr, "Failed to read data: %s\n", geterror());
			return PPSDK_EGENERAL;
		}
	}
	while (!(*nbr));	

	return PPSDK_ENOERR;
}

//-----------------------------------------------------------------------------
static unsigned int write(void *param, const unsigned char in[], unsigned int inLen,
						  unsigned int *nbw)
{
	BOOL res = rs232_write((HANDLE)param, in, inLen, nbw);

	if (res) {
		if ((*nbw) > 0) {
			if (VERBOSE) {
				printf(">> write(%d): ", *nbw); printb(in, *nbw); printf("\n");
			}
		}
	} else {
		fprintf(stderr, "Failed to write data: %s\n", geterror());
		return PPSDK_EGENERAL;
	}

	return PPSDK_ENOERR;	
}

//-----------------------------------------------------------------------------
static void emv2_online_authorize(void *handler, void* param, const unsigned char data[],
								  unsigned int dataLen)
{
	printf("emv2_online_authorize: ");
	printb(data, dataLen);
	printf("\n");

	{
		int cvmResult = bertlv_find_value_num(data, dataLen, 0xC2, -1);
		
		if (cvmResult == PPSDK_EMV2_CVM_ONLINE_PIN) {
			printf("CVM Result: ONLINE PIN\n");

			// Get ISO0 pin block
			// ppsdk_uiVerifyPINOnline
		} else if (cvmResult == PPSDK_EMV2_CVM_CONFIG_CODE_VERIFIED) {
			printf("CVM Result: CONFIG CODE VERIFIED\n");
		} else if (cvmResult == PPSDK_EMV2_CVM_OBTAIN_SIGNATURE) {
			printf("CVM Result: OBTAIN SIGNATURE\n");
		} else if (cvmResult == PPSDK_EMV2_CVM_NO_CVM) {
			printf("CVM Result: NO CVM\n");
		}  else if (cvmResult == PPSDK_EMV2_CVM_OFFLINE_PIN) {
			printf("CVM Result: OFFLINE PIN\n");
		}
	}

	{
		int paymentInterface = bertlv_find_value_num(data, dataLen, 0xC3, -1);

		if (paymentInterface == PPSDK_EMV2_INTERFACE_CONTACT) {
			printf("Payment Interface: CONTACT\n");
		} else if (paymentInterface == PPSDK_EMV2_INTERFACE_CONTACTLESS) {
			printf("Payment Interface: CONTACTLESS\n");
		} else if (paymentInterface == PPSDK_EMV2_INTERFACE_MAGNETIC) {
			printf("Payment Interface: INTERFACE_MAGNETIC\n");
		} else if (paymentInterface == PPSDK_EMV2_INTERFACE_MANUAL) {
			printf("Payment Interface: MANUAL\n");
		}
	}

	printf("Set online result\n");
	{
		unsigned char buffer[] = { 
			// C2 tag that specify whether online request is successful or not
			0xC2, 0x01, 0x01,

			// E6 is constructed tag that contains at least authorization response code 8A
			0xE6, 0x04, 0x8A, 0x02, 0x30, 0x30,
		};
		
		ppsdk_emv2SetOnlineResult(handler, buffer, sizeof(buffer));
	}
}
//-----------------------------------------------------------------------------
static void emv2_update_ui(void *handler, void* param, int message, int status, int hold)
{
	printf("emv2_update_ui: ");

	switch (message) {
        // EMV2 message codes
        case PPSDK_EMV2_MESSAGE_NOT_WORKING:
            printf("NOT WORKING\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_APPROVED:
            printf("APPROVED\n");
			break;
        case PPSDK_EMV2_MESSAGE_DECLINED:
            printf("DECLINED\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_PLEASE_ENTER_PIN:
            printf("PLEASE ENTER PIN\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_ERROR_PROCESSING:
            printf("ERROR PROCESSING\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_REMOVE_CARD:
            printf("REMOVE CARD\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_PRESENT_CARD:
            printf("PRESENT CARD\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_IDLE:
            printf("IDLE\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_PROCESSING:
            printf("PROCESSING\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_CARD_READ_OK_REMOVE:
            printf("CARD READ OK REMOVE\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_TRY_OTHER_INTERFACE:
            printf("TRY OTHER INTERFACE\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_CARD_COLLISION:
            printf("CARD COLLISION\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_SIGN_APPROVED:
            printf("SIGN APPROVED\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_ONLINE_AUTHORISATION:
            printf("ONLINE AUTHORISATION\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_TRY_OTHER_CARD:
            printf("TRY OTHER CARD\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_INSERT_CARD:
            printf("INSERT CARD\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_CLEAR_DISPLAY:
            printf("CLEAR DISPLAY\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_SEE_PHONE:
            printf("SEE PHONE\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_PRESENT_CARD_AGAIN:
            printf("PRESENT CARD AGAIN\n"); 
			break;
        case PPSDK_EMV2_MESSAGE_NA:
            printf("N/A\n"); 
			break;
        default:
            printf("N/A %d\n", message);
    }

}
//-----------------------------------------------------------------------------
static void emv2_complete(void *handler, void* param, const unsigned char data[], 
						  unsigned int dataLen)
{
	printf("emv2_complete: ");
	printb(data, dataLen);
	printf("\n");

	isTransactionActive = FALSE;
}

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Transaction implementation
///////////////////////////////////////////////////////////////////////////////
static BOOL initPinpad(void *handle, void **ref)
{
	void *pinpad = NULL;
	ppsdk_config config;	
	ppsdk_info info;
	ppsdk_date date;
	time_t t;

	memset(&config, 0, sizeof(ppsdk_config));
	config.read = read;
	config.write = write;	
	config.emv2_online_authorize = emv2_online_authorize;
	config.emv2_update_ui = emv2_update_ui;
	config.emv2_complete = emv2_complete;	

	CHECK_RESULT(ppsdk_init(&config, handle, ref));
	pinpad = *ref;

	CHECK_RESULT(ppsdk_getInfo(pinpad, &info));	
	printf("Pinpad information\n");
	printf("----------------------------------------------------------------------\n");
	printf("  Type: %X\n", info.devtype);
	printf("  Version: %d.%d.%d.%d\n", (unsigned char)(info.version >> 24), (unsigned char)(info.version >> 16), (unsigned char)(info.version >> 8), (unsigned char)info.version);
	printf("  Name: %s\n", info.name);
	printf("  INNOVA serial: "); printb(info.iser, sizeof(info.iser)); printf("\n"); 
	printf("  DATECS serial: %d\n", info.dser);
	printf("  Key loader version: %d.%d.%d.%d\n", (unsigned char)(info.klver >> 24), (unsigned char)(info.klver >> 16), (unsigned char)(info.klver >> 8), (unsigned char)info.klver);
	printf("----------------------------------------------------------------------\n");
	
	printf("Set time\n");		
	t = time(NULL);
	date.year = localtime(&t)->tm_year - 2000;
	date.month = localtime(&t)->tm_mon;
	date.day = localtime(&t)->tm_mday;
	date.hour = localtime(&t)->tm_hour;
	date.minute = localtime(&t)->tm_min;
	date.second = localtime(&t)->tm_sec;	
	CHECK_RESULT(ppsdk_sysSetDate(pinpad, &date));
	
	printf("Init display\n");
	CHECK_RESULT(ppsdk_uiInitScreen(pinpad));
	CHECK_RESULT(ppsdk_uiReleaseKeyboardControl(pinpad));
	CHECK_RESULT(ppsdk_uiOpenTextWindow(pinpad, 0, 0, 16, 2, PPSDK_FONT_8X16, PPSDK_CODETABLE_LATIN1));
				
	// Disable all events.
	printf("Disable events\n");
	CHECK_RESULT(ppsdk_sysEnableEvents(pinpad, 0));
				
	printf("Pinpad is ready.\n");

	return TRUE;
}
//-----------------------------------------------------------------------------
static BOOL loadKeys(void *pinpad)
{
	static unsigned char masterKeyBlock[] = { 0x42, 0x30, 0x30, 0x38, 0x30, 0x4B, 0x31, 0x54, 0x58, 0x30, 0x31, 0x45, 0x30, 0x30, 0x30, 0x30, 0x32, 0x32, 0x39, 0x34, 0x30, 0x32, 0x33, 0x41, 0x42, 0x42, 0x45, 0x33, 0x31, 0x31, 0x43, 0x44, 0x38, 0x34, 0x31, 0x46, 0x38, 0x45, 0x34, 0x43, 0x31, 0x39, 0x32, 0x44, 0x42, 0x35, 0x33, 0x43, 0x42, 0x33, 0x33, 0x43, 0x42, 0x37, 0x39, 0x36, 0x38, 0x31, 0x39, 0x34, 0x31, 0x36, 0x42, 0x30, 0x32, 0x41, 0x41, 0x38, 0x34, 0x34, 0x34, 0x33, 0x30, 0x30, 0x34, 0x42, 0x41, 0x30, 0x35, 0x39, 0 };
	static unsigned char dataKeyBlock[] = { 0x42, 0x30, 0x30, 0x38, 0x30, 0x44, 0x30, 0x54, 0x4E, 0x30, 0x30, 0x45, 0x30, 0x30, 0x30, 0x30, 0x37, 0x44, 0x41, 0x39, 0x42, 0x30, 0x37, 0x43, 0x35, 0x32, 0x44, 0x39, 0x35, 0x45, 0x43, 0x41, 0x45, 0x44, 0x43, 0x38, 0x46, 0x32, 0x34, 0x45, 0x41, 0x45, 0x45, 0x44, 0x45, 0x44, 0x32, 0x37, 0x35, 0x46, 0x38, 0x34, 0x41, 0x45, 0x30, 0x32, 0x35, 0x39, 0x35, 0x46, 0x34, 0x36, 0x32, 0x45, 0x39, 0x38, 0x45, 0x37, 0x39, 0x36, 0x44, 0x41, 0x33, 0x39, 0x45, 0x44, 0x34, 0x37, 0x35, 0x30, 0 };
	static unsigned char pinKeyBlock[] = { 0x42, 0x30, 0x31, 0x30, 0x34, 0x42, 0x31, 0x54, 0x58, 0x30, 0x30, 0x45, 0x30, 0x31, 0x30, 0x30, 0x32, 0x30, 0x31, 0x38, 0x46, 0x46, 0x46, 0x46, 0x30, 0x30, 0x30, 0x30, 0x30, 0x34, 0x30, 0x30, 0x30, 0x42, 0x32, 0x30, 0x30, 0x30, 0x30, 0x30, 0x43, 0x44, 0x38, 0x38, 0x46, 0x37, 0x31, 0x43, 0x45, 0x39, 0x44, 0x33, 0x46, 0x38, 0x32, 0x35, 0x38, 0x37, 0x45, 0x34, 0x37, 0x36, 0x30, 0x35, 0x32, 0x46, 0x35, 0x37, 0x42, 0x43, 0x42, 0x39, 0x37, 0x42, 0x35, 0x46, 0x41, 0x33, 0x45, 0x38, 0x39, 0x31, 0x44, 0x38, 0x35, 0x30, 0x42, 0x46, 0x44, 0x34, 0x41, 0x32, 0x37, 0x36, 0x43, 0x36, 0x33, 0x32, 0x35, 0x36, 0x41, 0x31, 0x36, 0x32, 0 };
	unsigned int res;

	printf("Load TR31 Master Key\n");

	res = ppsdk_cryptoExchangeCBCKey(pinpad, TMK_KEY_INDEX, TR31_MASTER_KEY_INDEX, PPSDK_KEY_TR31, 1, masterKeyBlock, NULL);
	if (res == PPSDK_EDUBKEY) {
		printf("TR31 Master Key is already loaded\n");
	} else {
		CHECK_RESULT(res);
	}

	printf("Load Data Key\n");
	res = ppsdk_cryptoExchangeCBCKey(pinpad, TR31_MASTER_KEY_INDEX, DATA_KEY_INDEX, PPSDK_KEY_TR31, 1, dataKeyBlock, NULL);
	if (res == PPSDK_EDUBKEY) {
		printf("Data Key is already loaded\n");
	} else {
		CHECK_RESULT(res);
	}

	printf("Load PIN Key\n");
	res = ppsdk_cryptoExchangeCBCKey(pinpad, TR31_MASTER_KEY_INDEX, PIN_KEY_INDEX, PPSDK_KEY_TR31, 1, pinKeyBlock, NULL);
	if (res == PPSDK_EDUBKEY) {
		printf("PIN Key is already loaded\n");
	} else {
		CHECK_RESULT(res);
	}

    printf("Save keys to NVRAM\n");
    CHECK_RESULT(ppsdk_cryptoSaveKeysToFlash(pinpad));
	
	return TRUE;
}
//-----------------------------------------------------------------------------
static BOOL loadConfiguration(void *pinpad)
{
	printf("Initialize Payment Engine\n");
    CHECK_RESULT(ppsdk_emv2Initialize(pinpad));

    printf("Load contactless configuration data\n");
    CHECK_RESULT(ppsdk_emv2LoadContactlessConfiguration(pinpad, CL_DATA_CFG, sizeof(CL_DATA_CFG)));

    printf("Load contactless CA keys\n");    
    CHECK_RESULT(ppsdk_emv2LoadContactlessCAPK(pinpad, CL_CA_KEYS_CFG, sizeof(CL_CA_KEYS_CFG)));

    printf("Load contact configuration data\n");
    CHECK_RESULT(ppsdk_emv2LoadContactConfiguration(pinpad, CHIP_DATA_CFG, sizeof(CHIP_DATA_CFG)));
    
    printf("Load contact CA keys\n");
    CHECK_RESULT(ppsdk_emv2LoadContactCAPK(pinpad, CHIP_CA_KEYS_CFG, sizeof(CHIP_CA_KEYS_CFG)));

    printf("Deinitialise Payment Engine\n");
    CHECK_RESULT(ppsdk_emv2Deinitialise(pinpad));

	return TRUE;
}
//-----------------------------------------------------------------------------
static BOOL processTransaction(void *pinpad, int timeout)
{	
	unsigned char buffer[64];
	unsigned char *ptr;
	DWORD complete;

	// Prepare initial transaction data
	ptr = buffer;	
	ptr += bertlv_encode_num(ptr, TAG_9C_TRANSACTION_TYPE, 0x00, 1);	
	ptr += bertlv_encode_num(ptr, TAG_9F02_AMOUNT_AUTHORISED_NUM, 100, 12);	
	ptr += bertlv_encode_num(ptr, TAG_5F2A_TRANSACTION_CURR_CODE, 840, 3);

	printf("Initialize Payment Engine\n");
    CHECK_RESULT(ppsdk_emv2Initialize(pinpad));

	printf("Lock keyboard\n");
	CHECK_RESULT(ppsdk_uiGetKeyboardControl(pinpad));
    
	printf("Start transaction\n");
    CHECK_RESULT(ppsdk_emv2StartTransaction(pinpad, 
		    (PPSDK_EMV2_INTERFACE_CONTACT | PPSDK_EMV2_INTERFACE_CONTACTLESS |
			 PPSDK_EMV2_INTERFACE_MAGNETIC | PPSDK_EMV2_INTERFACE_MANUAL),
            0 /* Flags */,
            buffer, ptr - buffer,
            timeout / 1000 /* Seconds */));

	isTransactionActive = TRUE;

	complete = GetTickCount() + timeout + 1000;
	do {
		// Call it to process pinpad events.
		int res = ppsdk_pull(pinpad, 50);

		if (res != PPSDK_ETIMEOUT)
		{
			CHECK_RESULT(res);
		}

		// Stop transaction if it takes too long
		if ((long)(complete) - (long)(GetTickCount()) < 0) {
			CHECK_RESULT(res); // TIMEOUT
		}
	} while (isTransactionActive);

    printf("Deinitialise Payment Engine\n");
    CHECK_RESULT(ppsdk_emv2Deinitialise(pinpad));

	return TRUE;
}
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Application entry
///////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
	char port[MAX_PATH];
	HANDLE handle = INVALID_HANDLE_VALUE;	
	void *pinpad = NULL;
	BOOL result = TRUE;
	
	if (argc > 1) {
		strcpy(port, argv[1]);
	} else {
		strcpy(port, DEFAULT_SERIAL_PORT);
	}

	// Open serial port connection.
	printf("Open serial port\n");
	handle = rs232_open(port, 115200, 8, PARITY_NONE, TWOSTOPBITS);
	if (handle == INVALID_HANDLE_VALUE) {
		fprintf(stderr, "Failed to open port: %s\n", geterror());
		result = FALSE;
	} else {
		rs232_purge(handle);
	}	
	
	if (result) {		
		result = initPinpad(handle, &pinpad);
	}
		
	// Pinpad keys are stored in NVRAM and need to be reaload only when update.
	if (result) {		
		result = loadKeys(pinpad);
	}
	
	// Pinpad configuration is stored in NVRAM and need to be reaload only when update.
	if (result) {
		result = loadConfiguration(pinpad);
	}
	
	if (result) {
		result = processTransaction(pinpad, MAX_TRANSACTION_TIME);
	}

	// Release resources
	if (pinpad != NULL) {
		printf("Release pinpad\n");
		ppsdk_release(pinpad);
	}

	if (handle != INVALID_HANDLE_VALUE) {
		printf("Close serial port\n");
		rs232_close(handle);
	}
	
	printf("Press ENTER to exit...");
	getchar();
	return 0;
}
///////////////////////////////////////////////////////////////////////////////
