/**
 * BER-TLV data object implementation (ISO 7816 Annex D.1)
 * 
 * Each BER-TLV data object (see ISO/IEC 8825) shall consist of 2 or 3 
 * consecutive fields : 
 *  - The tag field (T) consists of one or more consecutive bytes. It encodes a 
 *    class, a type and a number.
 *  - The length field consists of one or more consecutive bytes. It encodes 
 *    an integer (L).
 *  - If (L) is not null, then the value field (V) consists of (L) consecutive
 *    bytes. If L is null, then the data object is empty: there is no value field.
 *
 * (C) Copyright (C) 2015, Datecs Ltd.
 * 
 * No part of this software may be used, stored, compiled, reproduced,
 * modified, transcribed, translated, transmitted, or transferred, in any form
 * or by any means whether electronic, mechanical, magnetic, optical,
 * or otherwise, without the express prior written permission.  
 */

#ifndef _BERTLV_H
#define _BERTLV_H

// Types
typedef unsigned char   bertlv_u8;
typedef unsigned int    bertlv_u32;

// BER-TLV error
#define BERTLV_ERROR   0   /* Error occurs during data processing. */

// BER-TLV tag types
#define BERTLV_TAG_TYPE_PRIMITIVE    0  /* A primitive data object where the value field contains
                                         * a data element for financial transaction interchange. 
                                         */
#define BERTLV_TAG_TYPE_CONSTRUCTED  1  /* A constructed data object where the value field contains one or more
                                         * primitive or constructed data objects. The value field of a constructed
                                         * data object is called a template.
                                         */
// BER-TLV tag class types
#define BERTLV_TAG_CLASS_UNIVERSAL        0  /* The type is native to ASN.1 */
#define BERTLV_TAG_CLASS_APPLICATION      1  /* The type is only valid for one specific application. */
#define BERTLV_TAG_CLASS_CONTEXT_SPECIFIC 2  /* Meaning of this type depends on the context (such as within a sequence, set or choice). */
#define BERTLV_TAG_CLASS_PRIVATE          3  /* Defined in private specifications. */

/**
 * Get whether tag value is valid. 
 *
 * @param tag a tag value.
 *
 * @return size of encoded tag in bytes or 0 if tag is invalid.
 */
bertlv_u32 bertlv_is_tag_valid(bertlv_u32 tag);

/**
 * Get tag type.
 *
 * @param tag a tag value.
 *
 * @return tag type, one of BERTLV_TAG_TYPE_* constant values.
 */
bertlv_u32 bertlv_get_tag_type(bertlv_u32 tag);

/**
 * Get tag class.
 *
 * @param tag a tag value.
 *
 * @return tag class, one of BERTLV_TAG_CLASS_* constant values.
 */
bertlv_u32 bertlv_get_tag_class(bertlv_u32 tag);

/**
 * Encode TLV tag.
 * 
 * @param buffer a buffer used to store encoded tag value.
 *               This parameter can be NULL. 
 * @param tag a tag value.
 *
 * @return size of encoded tag in bytes.
 */
bertlv_u32 bertlv_encode_tag(bertlv_u8 *buffer, bertlv_u32 tag);

/**
 * Encode TLV length.
 * 
 * @param buffer a buffer used to store encoded length value. 
 *               This parameter can be NULL. 
 * @param length a length value. 
 *
 * @return size of encoded length.
 */
bertlv_u32 bertlv_encode_length(bertlv_u8 *buffer, bertlv_u32 length);

/**
 * Encode TLV tag and length.
 * 
 * @param buffer a buffer used to store encoded length value.
 *               This parameter can be NULL. 
 * @param tag a tag value. 
 * @param length a length value. 
 *
 * @return size of encoded length.
 */
bertlv_u32 bertlv_encode_tag_length(bertlv_u8 *buffer, bertlv_u32 tag, bertlv_u32 length);

/**
 * Encode tag and data into TLV.
 *
 * @param buffer a buffer used to store encoded TLV. This parameter can be NULL.
 * @param tag a tag value.
 * @param length a length value (or length of parameter 'value' in bytes). 
 * @param value a pointer to buffer that contains value data.
 *
 * @return size of encoded TLV in bytes or 0 on error.
 */
bertlv_u32 bertlv_encode(bertlv_u8 *buffer, bertlv_u32 tag, bertlv_u32 length,
                         const bertlv_u8 *value);

// ----------------------------------------------------------------------------
// 
// ----------------------------------------------------------------------------
/**
 * Encode integer with two decimal digits per byte and left-padded with zeros.
 *
 * @param buffer a buffer used to store encoded TLV. This parameter can be NULL.
 * @param tag a tag value.
 * @param value a value data.
 * @param decimals number of digits to encode.
 *
 * @return size of encoded TLV in bytes or 0 on error.
 */
bertlv_u32 bertlv_encode_num(bertlv_u8 *buffer, bertlv_u32 tag, 
							 bertlv_u32 value, bertlv_u8 decimals);

/**
 * Decode tag from buffer.
 *
 * @param buffer a buffer that contains the encoded tag.
 * @param count the number of bytes in buffer.
 * @param tag a pointer to variable that receives decoded tag.  
 *            This parameter can be NULL. 
 *
 * @return size of encoded tag in bytes or 0 on error.
 */
bertlv_u32 bertlv_decode_tag(const bertlv_u8 *buffer, bertlv_u32 count, 
							 bertlv_u32 *tag);

/**
 * Decode length from buffer.
 *
 * @param buffer a buffer that contains the encoded length.
 * @param count the number of bytes in buffer.
 * @param length a pointer to variable that receives decoded length. 
 *               This parameter can be NULL. 
 *
 * @return size of encoded length in bytes or 0 on error.
 */
bertlv_u32 bertlv_decode_length(const bertlv_u8 *buffer, bertlv_u32 count, 
								bertlv_u32 *length);

/**
 * Decode TLV from buffer.
 *
 * @param buffer a buffer that contains the encoded TLV.
 * @param count the number of bytes in buffer.
 * @param tag a pointer to variable that receives decoded tag. 
 *            This parameter can be NULL. 
 * @param length a pointer to variable that receives decoded length. 
 *               This parameter can be NULL.
 * @param value a pointer to variable that receives address of value. 
 *             This parameter can be NULL. 
 *
 * @return size of encoded TLV in bytes or 0 on error.
 */
bertlv_u32 bertlv_decode(const bertlv_u8 *buffer, bertlv_u32 count, 
						 bertlv_u32 *tag, bertlv_u32 *length,
                         const bertlv_u8 **value);

/**
 * Find a TLV by given tag value.
 *
 * @param buffer a pointer to buffer that contains encoded TLVs.
 * @param count the size of buffer in bytes.
 * @param tag the tag value.
 * @param ptr a pointer to variable that receives address of TLV. 
 *
 * @return size of encoded TLV in bytes or 0 if is not found.
 */
bertlv_u32 bertlv_find(const bertlv_u8 *buffer, bertlv_u32 count, 
                       bertlv_u32 tag, const bertlv_u8 **ptr);


/**
 * Find a TLV by given tag and return its value as numeric.
 *
 * @param buffer a pointer to buffer that contains encoded TLVs.
 * @param count the size of buffer in bytes.
 * @param tag the tag value.
 * @param defValue default value if tag is not found.
 *
 * @return the numeric TLV value.
 */
bertlv_u32 bertlv_find_value_num(const bertlv_u8 *buffer, 
								   bertlv_u32 count,
								   bertlv_u32 tag, 									    
								   bertlv_u32 defValue);

#endif // _BERTLV_H
