/**
 * ISO8583 message protocol implementation.
 *
 * (C) Copyright (C) 2012, Datecs Ltd.
 *
 * No part of this software may be used, stored, compiled, reproduced,
 * modified, transcribed, translated, transmitted, or transferred, in any form
 * or by any means whether electronic, mechanical, magnetic, optical,
 * or otherwise, without the express prior written permission.
 */

#ifndef _CRT_SECURE_NO_DEPRECATE
#define _CRT_SECURE_NO_DEPRECATE 1
#endif

#include <stdlib.h>
#include <string.h>

#include "bertlv.h"
#include "pinpad.h"

// Default timeout for I/O operation.
#define	DEFAULT_TIMEOUT	5000

// Packet header size
#define	PACKET_HEADER_SIZE 5

// Pinpad I/O buffer size
#define	PACKET_BUFFER_SIZE 2048

// Check result and return on error.
#define CHECK_RESULT(X) do { unsigned int r = (X); if (r != PPSDK_ENOERR) return r; } while (0);

// Card inserted
#define EMV2_IFSTAT_INSERTED          1
// Card removed
#define EMV_IFSTAT_INTERFACE_REMOVED  2
// Interface error
#define EMV_IFSTAT_INTERFACE_ERROR    3

////////////////////////////////////////////////////////////////////////////////
// Local declarations
////////////////////////////////////////////////////////////////////////////////
// Handler context structure
typedef struct {	
	ppsdk_config config; // Configuration structure.
	unsigned int status;    // The status code from last executed command.
	unsigned char *packet;    // The buffer, used to cache I/O data.
	unsigned int offset;	 // The buffer offset. 	
	void *param;
} hcontext;
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Local functions
////////////////////////////////////////////////////////////////////////////////
// Helper method to read the exact amount of data.
static PPSDK_ERR readFully(const hcontext *context, unsigned char out[], unsigned int numberOfBytesToRead, unsigned int timeout) 
{		
	unsigned int nbr = 0;

	while (numberOfBytesToRead > 0) {
		CHECK_RESULT(context->config.read(context->param, out, numberOfBytesToRead, &nbr, timeout));
		numberOfBytesToRead -= nbr;		
		out += nbr;
	} 

	return PPSDK_ENOERR;
}
//-----------------------------------------------------------------------------
// Helper method to write the exact amount of data.
static PPSDK_ERR writeFully(const hcontext *context, const unsigned char in[], unsigned int numberOfBytesToWrite) 
{	
	unsigned int nbw = 0;

	while (numberOfBytesToWrite > 0) {		
		CHECK_RESULT(context->config.write(context->param, in, numberOfBytesToWrite, &nbw));		
		numberOfBytesToWrite -= nbw;
		in += nbw;
	} 
		
	return PPSDK_ENOERR;
}
//-----------------------------------------------------------------------------
// Append byte to packet buffer
static PPSDK_ERR writePacketByte(hcontext *context, unsigned char b)
{
	if (context->offset < PACKET_BUFFER_SIZE) {
		context->packet[context->offset++] = b;		
		return PPSDK_ENOERR;
	} else {
		return PPSDK_EINVLEN;
	}
}
//-----------------------------------------------------------------------------
// Append data to packet buffer
static PPSDK_ERR writePacketData(hcontext *context, const unsigned char in[], unsigned int inLen)
{
	unsigned int n = 0;

	if (in == NULL && inLen > 0) return PPSDK_EINVPAR;

	for (n = 0; n < inLen; n++) {
		CHECK_RESULT(writePacketByte(context, in[n]));
	}

	return PPSDK_ENOERR;
}
//-----------------------------------------------------------------------------
// Initialize packet buffer
static PPSDK_ERR constructPacket(hcontext *context, unsigned char command)
{	
	if (context == NULL) return PPSDK_EINVPAR;

	context->offset = 0;
	CHECK_RESULT(writePacketByte(context, 0x3E));
	CHECK_RESULT(writePacketByte(context, command));	
	CHECK_RESULT(writePacketByte(context, 0));	
	CHECK_RESULT(writePacketByte(context, 0));	
	CHECK_RESULT(writePacketByte(context, 0));	

	return PPSDK_ENOERR;
}
//-----------------------------------------------------------------------------
// Write packet
static PPSDK_ERR writePacket(hcontext *context)
{
	unsigned int len = context->offset - PACKET_HEADER_SIZE; // data size
	unsigned char crc = 0;
	unsigned int n = 0;

	// Set packet data length
	context->packet[PACKET_HEADER_SIZE - 2] = (unsigned char)(len >> 8);
	context->packet[PACKET_HEADER_SIZE - 1] = (unsigned char)(len);
		
	// Calculate CRC
	for (n = 0; n < context->offset; n++) {
		crc ^= context->packet[n];
	}

	CHECK_RESULT(writePacketByte(context, crc));
	CHECK_RESULT(writeFully(context, context->packet, context->offset));

	return PPSDK_ENOERR;
}
//-----------------------------------------------------------------------------
// Read packet
static PPSDK_ERR readPacket(hcontext *context, unsigned int timeout)
{
	unsigned int len = 0;
	unsigned char crc = 0;
	unsigned int n = 0;
	unsigned char evn = 0;
	unsigned char ok = 0;

	do {
		// Receive command data
		CHECK_RESULT(readFully(context, context->packet, PACKET_HEADER_SIZE, timeout));
		if (context->packet[0] != 0x3E) {
			return PPSDK_EGENERAL;
		}

		len = (context->packet[3] << 8) + (context->packet[4]);
		CHECK_RESULT(readFully(context, context->packet + PACKET_HEADER_SIZE, len, DEFAULT_TIMEOUT));	
		CHECK_RESULT(readFully(context, &crc, 1, DEFAULT_TIMEOUT));

		// Calculate CRC
		for (n = 0; n < (PACKET_HEADER_SIZE + len); n++) {
			crc ^= context->packet[n];
		}

		if (crc != 0) {
			return PPSDK_EINVCRC;
		}
		
		// Check for event
		evn = context->packet[1];
		if (evn != 0) {
			unsigned char *ptr = &context->packet[5];			
			unsigned int psize = len;			

			// Common event processing.
			if (context->config.notify != NULL) {
				context->config.notify(context, context->param, evn, ptr, psize);
			}
			
			// Process EMV events
			if (evn == PPSDK_EVENT_EMV2) {				
				switch (ptr[0]) {
					case PPSDK_EMV2_EVENT_ONLINE_AUTHORIZATION: {
						if (context->config.emv2_online_authorize != NULL) {
							context->config.emv2_online_authorize(context, context->param, ptr + 2, len - 2);
						}
						break;
					}
				    case PPSDK_EMV2_EVENT_UPDATE_UI: {						
						if (context->config.emv2_update_ui != NULL) {
							int message = 0;
							int status = -1;
							int hold = -1;
	                
							if ((len - 2) > 3)
							{
								message = ((ptr[2 + 2] & 0xff) << 8) + (ptr[2 + 3] & 0xff);
							}

							if ((len - 2) > 9)
							{
								status = (ptr[2 + 9] & 0xff);
							}

							if ((len - 2) > 12)
							{
								hold = ((ptr[2 + 11] & 0xff) << 8) + (ptr[2 + 12] & 0xff);
							}  

							context->config.emv2_update_ui(context, context->param, message, status, hold);
						}

						break;
					}
				    case PPSDK_EMV2_EVENT_COMPLETE: {
						if (context->config.emv2_complete != NULL) {
							context->config.emv2_complete(context, context->param, ptr + 2, len - 2);
						}
						break;
					}
				    case PPSDK_EMV2_EVENT_INTERFACE: {
						int inter = ptr[2 + 0];
						int error = ptr[2 + 1];

						if (inter == 0) {
							if(error == EMV_IFSTAT_INTERFACE_ERROR) {
								ppsdk_emv2CancelTransaction(context);

								if (context->config.emv2_complete != NULL) {
									unsigned char buffer[] = {
										0xC1, 0x02, 0x01, 0x00,
										0xC3, 0x01, 0x01,
										0xC4, 0x01, 99,
									};

									context->config.emv2_complete(context, context->param, buffer, sizeof(buffer));
								}								
							}
						}
						break;
					}
				}				
			}

			continue;
		}

		ok = 1;
	} while (!ok);

	context->offset = PACKET_HEADER_SIZE + len;
	context->status = context->packet[2];
	
	return (PPSDK_ERR)context->status;
}
//-----------------------------------------------------------------------------
// Transmit packet buffer
static PPSDK_ERR transmitPacket(hcontext *context, unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	unsigned int len = 0;

	CHECK_RESULT(writePacket(context));
	CHECK_RESULT(readPacket(context, DEFAULT_TIMEOUT));
	len = context->offset - PACKET_HEADER_SIZE;

	if (out != NULL) {
		if (len > outMaxLen)
			return PPSDK_EINVLEN;
		 
		memcpy(out, context->packet + PACKET_HEADER_SIZE, len);
	}

	if (nbr) {
		*nbr = len;
	}
		
	return PPSDK_ENOERR;
}
//-----------------------------------------------------------------------------
// EMV transmit packet buffer.
static PPSDK_ERR emvTransmitPacket(hcontext *context, unsigned char out[], unsigned int outMaxLen, unsigned int *nbr, unsigned short *sw)
{
	unsigned char *p = context->packet + PACKET_HEADER_SIZE;
	unsigned int len = 0;
	
	CHECK_RESULT(writePacket(context));
	CHECK_RESULT(readPacket(context, DEFAULT_TIMEOUT));
	len = context->offset - PACKET_HEADER_SIZE;

	if (len < 2) return PPSDK_EINVLEN;

	if (sw != NULL) {		
		*sw = (unsigned short)(p[0] + (p[1] << 8));
	}

	p+= 2;
	len -= 2;

	if (out != NULL) {
		if (len > outMaxLen)
			return PPSDK_EINVLEN;
		 
		memcpy(out, p, len);
	}

	if (nbr) {
		*nbr = len;
	}
		
	return PPSDK_ENOERR;
}
//-----------------------------------------------------------------------------
// Helper method to convert from byte array to int in using little endian
static unsigned int byteArrayToInt(const unsigned char buf[], unsigned char len)
{
	unsigned char n = 0;
	unsigned int res = 0;

	for (n = 0; n < len; n++) {
		res = res + (buf[n] << (8 * n));
	}
	
	return res;
}
//-----------------------------------------------------------------------------
// Helper method to convert from byte array to int in using little endian
static unsigned int byteArrayToIntBE(const unsigned char buf[], unsigned char len)
{
	unsigned char n = 0;
	unsigned int res = 0;

	for (n = 0; n < len; n++) {
		res = (res << 8) + buf[n];
	}
	
	return res;
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// General functions
////////////////////////////////////////////////////////////////////////////////
PPSDK_ERR ppsdk_init(const ppsdk_config *config, void *param, void **handler)
{
	hcontext *context = NULL;
	unsigned int res = PPSDK_ENOERR;

	if (config == NULL) 
		return PPSDK_EINVPAR;

	if (config->read == NULL) 
		return PPSDK_EINVPAR;

	if (config->write == NULL) 
		return PPSDK_EINVPAR;

	if (handler == NULL) 
		return PPSDK_EINVPAR;

	context = (hcontext *)malloc(sizeof(hcontext));
	if (context == NULL)
		return PPSDK_EGENERAL;	
	
	memset(context, 0, sizeof(context));
	memcpy(&context->config, config, sizeof(ppsdk_config));		
	context->param = param;

	context->packet = (unsigned char *)malloc(PACKET_BUFFER_SIZE);
	if (context->packet == NULL) {
		ppsdk_release(context);
		return PPSDK_EGENERAL;	
	}

	res = ppsdk_getInfo(context, NULL);
	if (res != PPSDK_ENOERR) {
		ppsdk_release(context);
		return res;
	}

	*handler = context;
	return PPSDK_ENOERR;
}

//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_release(void *handler)
{
	hcontext *context = (hcontext *)handler;

	if (context != NULL) {
		if (context->packet != NULL) {
			free(context->packet);
		}
		free(context);
	}

	return PPSDK_ENOERR;
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_transmit(void *handler, unsigned char command, 
						 const unsigned char in[], unsigned int inLen, 
                         unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	CHECK_RESULT(constructPacket((hcontext *)handler, command));	
	CHECK_RESULT(writePacketData((hcontext *)handler, in, inLen));

	return transmitPacket((hcontext *)handler, out, outMaxLen, nbr);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_pull(void *handler, unsigned int wait)
{
	hcontext *context = (hcontext *)handler;
	
	return readPacket(context, wait);
}
//-----------------------------------------------------------------------------
const char *ppsdk_getErrorText(unsigned int err)
{
	switch (err) {
		case PPSDK_ENOERR: return "No error";
		case PPSDK_EGENERAL: return "Some error";
		case PPSDK_EINVCMD: return "No valid command or subcommand code";
		case PPSDK_EINVPAR: return "Invalid paremeter";
		case PPSDK_EINVADR: return "Address is outside limits";
		case PPSDK_EINVVAL: return "Value is outside limits";
		case PPSDK_EINVLEN: return "Length is outside limits";
		case PPSDK_ENOPERMIT: return "The action is not permit in current state";
		case PPSDK_ENODATA: return "There is no data to be returned";
		case PPSDK_ETIMEOUT: return "Timeout is occur";
		case PPSDK_EKEYNUM: return "Invalid key number";
		case PPSDK_EKEYATTR: return "Invalid key attributes(usage)";		
		case PPSDK_EINVDEVICE: return "Calling of non-existing device";
		case PPSDK_ENOSUPPORT: return "(No used in this FW version)";
		case PPSDK_EPINLIMIT: return "Pin entering limit exceed";
		case PPSDK_EFLASH: return "Some error in flash commands";
		case PPSDK_EHARD: return "Some hardware error";
		case PPSDK_EINVCRC: return "Invalid checksum";
		case PPSDK_ECANCEL: return "The button 'CANCEL' is pressed";
		case PPSDK_EINVSIGN: return "Invalid signature";
		case PPSDK_EINVHEAD: return "Invalid data in header";
		case PPSDK_EINVPASS: return "Incorrent password";
		case PPSDK_EKEYFORMAT: return "Invalid key format";
		case PPSDK_ESCR: return "Some error in smart card reader";
		case PPSDK_EHAL: return "Error code is returned from HAL functions";
		case PPSDK_EINVKEY: return "Invalid key (may be no pressent)";
		case PPSDK_ENOPINDATA: return "The PIN length is <4 or >12";
		case PPSDK_EINVREMAINDER: return "Issuer or ICC key invalid remainder length";
		case PPSDK_ENOINIT: return "(No used in this FW version)";
		case PPSDK_ELIMIT: return "(No used in this FW version)";
		case PPSDK_EINVSEQ: return "(No used in this FW version)";
		case PPSDK_ENOPERM: return "The action is not permited";
		case PPSDK_ENOTMK: return "TMK is not loaded. The action cannot be executed";
		case PPSDK_EINVKEK: return "Wrong kek format";
		case PPSDK_EDUBKEY: return "Dublicated key";
		case PPSDK_EKBD: return "General keyboard error";
		case PPSDK_EKBDNOCAL: return " The keyboard is no calibrated";
		case PPSDK_EKBDBUG: return "Keyboard bug detected";
		case PPSDK_EBUSY: return "The device is busy, tray again";
		case PPSDK_ETAMPERED: return "The devicete in security tampered";
		case PPSDK_EEMSR: return "Error in encrypted head";
		case PPSDK_EACCEPT: return "The button 'OK' is pressed";
		default: return "Unknown error code";		
	}
}

////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// GET IDENTIFICATION INFO AND SERIAL NUMBER
////////////////////////////////////////////////////////////////////////////////
PPSDK_ERR ppsdk_getInfo(void *handler, ppsdk_info *info)
{	
	unsigned char tmp[64];
	unsigned int nbr = 0;
	
	CHECK_RESULT(ppsdk_transmit(handler, 0x01, NULL, 0, tmp, sizeof(tmp), &nbr));
	
	if (nbr < 45) { 
		return PPSDK_EINVLEN;
	}

	if (info != NULL) {	
		memset(info, 0, sizeof(ppsdk_info));
		info->devtype = byteArrayToInt(&tmp[0], 1);
		info->version = byteArrayToInt(&tmp[1], 4);
		memcpy(info->name, &tmp[ 5], 16);
		memcpy(info->iser, &tmp[21], 16);
		info->dser = byteArrayToInt(&tmp[37], 4);
		info->klver = byteArrayToInt(&tmp[41], 4);
	}

	return PPSDK_ENOERR;
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// 3DES KEYS MANAGMENT, CRYPTO, HASH
////////////////////////////////////////////////////////////////////////////////
PPSDK_ERR ppsdk_cryptoEncryptECB(void *handler, unsigned char key, const unsigned char in[], unsigned int inLen, unsigned char out[])
{
	hcontext *context = (hcontext *)handler;
	
	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x01));
	CHECK_RESULT(writePacketByte(context, 0x00));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketData(context, in, inLen));

	return transmitPacket(context, out, inLen, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoDecryptECB(void *handler, unsigned char key, const unsigned char in[], unsigned int inLen, unsigned char out[])
{
	hcontext *context = (hcontext *)handler;
	
	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x01));
	CHECK_RESULT(writePacketByte(context, 0x01));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketData(context, in, inLen));

	return transmitPacket(context, out, inLen, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoEncryptCBC(void *handler, unsigned char key, const unsigned char in[], unsigned int inLen, unsigned char out[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x02));
	CHECK_RESULT(writePacketByte(context, 0x00));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketData(context, in, inLen));

	return transmitPacket(context, out, inLen, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoDecryptCBC(void *handler, unsigned char key, const unsigned char in[], unsigned int inLen, unsigned char out[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x02));
	CHECK_RESULT(writePacketByte(context, 0x01));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketData(context, in, inLen));

	return transmitPacket(context, out, inLen, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoHash(void *handler, unsigned char type, const unsigned char in[], unsigned int inLen, unsigned char out[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x03));
	CHECK_RESULT(writePacketByte(context, type));
	CHECK_RESULT(writePacketData(context, in, inLen));

	return transmitPacket(context, out, 0xFF, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoSaveKeysToFlash(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x05));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoCrypt(void *handler, unsigned char mode, unsigned char key, const unsigned char iv[8], 
					  const unsigned char variant[16], const unsigned char *in, unsigned int inLen, unsigned char out[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x06));
	CHECK_RESULT(writePacketByte(context, mode));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketData(context, iv, 8));
	CHECK_RESULT(writePacketData(context, variant, 16));
	CHECK_RESULT(writePacketData(context, in, inLen));
	
	return transmitPacket(context, out, inLen, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoDeleteKey(void *handler, unsigned char key)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x09));
	CHECK_RESULT(writePacketByte(context, key));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoGetKCV(void *handler, unsigned char key, unsigned char kcv[8])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x0A));
	CHECK_RESULT(writePacketByte(context, key));
	
	return transmitPacket(context, kcv, 8, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoSetDUKPT(void *handler, unsigned char kek, unsigned char key, const unsigned char data[16])
{
	hcontext *context = (hcontext *)handler;
	
	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x0B));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketByte(context, kek));
	CHECK_RESULT(writePacketData(context, data, 16));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoCalculateMAC(void *handler, unsigned char mode, unsigned char key, const unsigned char iv[8],
							 const unsigned char variant[16], const unsigned char in[], unsigned int inLen, unsigned char mac[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x07));
	CHECK_RESULT(writePacketByte(context, mode));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketData(context, iv, 8));
	CHECK_RESULT(writePacketData(context, variant, 16));
	CHECK_RESULT(writePacketData(context, in, inLen));
	
	return transmitPacket(context, mac, 8, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoExchangeECBKey(void *handler, unsigned char kek, unsigned char key, unsigned char usage, 
							unsigned char version, const unsigned char data[16], unsigned char kcv[8])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x0D));
	CHECK_RESULT(writePacketByte(context, kek));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketByte(context, usage));
	if (usage == PPSDK_KEY_TR31) {
		CHECK_RESULT(writePacketByte(context, 0));
	} else {
		CHECK_RESULT(writePacketByte(context, version));
	}	
	CHECK_RESULT(writePacketByte(context, 0));
	if (usage == PPSDK_KEY_TR31) {
		CHECK_RESULT(writePacketData(context, data, strlen(data)));
		CHECK_RESULT(writePacketByte(context, 0));
	} else {
		CHECK_RESULT(writePacketData(context, data, 16));
	}	

	return transmitPacket(context, kcv, 8, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_cryptoExchangeCBCKey(void *handler, unsigned char kek, unsigned char key, unsigned char usage, 
							unsigned char version, const unsigned char data[16], unsigned char kcv[8])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x31));	
	CHECK_RESULT(writePacketByte(context, 0x0D));
	CHECK_RESULT(writePacketByte(context, kek));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketByte(context, usage));
	if (usage == PPSDK_KEY_TR31) {
		CHECK_RESULT(writePacketByte(context, 0));
	} else {
		CHECK_RESULT(writePacketByte(context, version));
	}	
	CHECK_RESULT(writePacketByte(context, 1));
	if (usage == PPSDK_KEY_TR31) {
		CHECK_RESULT(writePacketData(context, data, strlen(data)));
		CHECK_RESULT(writePacketByte(context, 0));
	} else {
		CHECK_RESULT(writePacketData(context, data, 16));
	}	

	return transmitPacket(context, kcv, 8, NULL);
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// CA KEYS MANAGMENT, EMV COMMANDS
////////////////////////////////////////////////////////////////////////////////
PPSDK_ERR ppsdk_caImportKeys(void *handler, unsigned char key, const unsigned char rid[], unsigned char idx, 
							 const unsigned char mod[], unsigned int modLen, 
							 const unsigned char exp[], unsigned int expLen)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x00));
	CHECK_RESULT(writePacketByte(context, key));	
	CHECK_RESULT(writePacketData(context, rid, 5));
	CHECK_RESULT(writePacketByte(context, idx));
	CHECK_RESULT(writePacketByte(context, expLen));
	CHECK_RESULT(writePacketByte(context, modLen));
	CHECK_RESULT(writePacketData(context, exp, expLen));
	CHECK_RESULT(writePacketData(context, mod, modLen));

	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_caWriteKeysToFlash(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x01));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_caGetCAKeyInfo(void *handler, unsigned char data[240])
{
	hcontext *context = (hcontext *)handler;
	
	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x02));

	return transmitPacket(context, data, 240, NULL);
	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_caImportIssuerKey(void *handler, unsigned char key, 
								  const unsigned char mod[], unsigned int modLen, 
								  const unsigned char exp[], unsigned int expLen, 
								  const unsigned char cert[], unsigned int certLen,
								  unsigned char out[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x03));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketByte(context, expLen));
	CHECK_RESULT(writePacketByte(context, modLen));
	CHECK_RESULT(writePacketByte(context, certLen));
	CHECK_RESULT(writePacketData(context, exp, expLen));
	CHECK_RESULT(writePacketData(context, mod, modLen));
	CHECK_RESULT(writePacketData(context, cert, certLen));
	
	return transmitPacket(context, out, certLen, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_caImportICCKey(void *handler, unsigned char type, 
								  const unsigned char mod[], unsigned int modLen, 
								  const unsigned char exp[], unsigned int expLen, 
								  const unsigned char cert[], unsigned int certLen,
								  unsigned char out[])
{
	hcontext *context = (hcontext *)handler;
	
	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x04));
	CHECK_RESULT(writePacketByte(context, type));
	CHECK_RESULT(writePacketByte(context, expLen));
	CHECK_RESULT(writePacketByte(context, modLen));
	CHECK_RESULT(writePacketByte(context, certLen));
	CHECK_RESULT(writePacketData(context, exp, expLen));
	CHECK_RESULT(writePacketData(context, mod, modLen));
	CHECK_RESULT(writePacketData(context, cert, certLen));
	
	return transmitPacket(context, out, certLen, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_caRSAVerify(void *handler, unsigned char type, 
							const unsigned char in[], unsigned int inLen, 
							unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;
	
	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x05));
	CHECK_RESULT(writePacketByte(context, type));
	CHECK_RESULT(writePacketByte(context, inLen));
	CHECK_RESULT(writePacketData(context, in, inLen));
	
	return transmitPacket(context, out, outMaxLen, nbr);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_dukptGenerateKey(void *handler, unsigned char key, unsigned char ksn[16])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x06));
	CHECK_RESULT(writePacketByte(context, key));	
	
	return transmitPacket(context, ksn, 16, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_dukptCalculateMAC(void *handler, const unsigned char iv[8], 
								  const unsigned char data[], unsigned int dataLen, unsigned char mac[8])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x07));
	CHECK_RESULT(writePacketData(context, iv, 8));	
	CHECK_RESULT(writePacketData(context, data, dataLen));	
	
	return transmitPacket(context, mac, 8, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_dukptGetCurrentKSN(void *handler, unsigned char ksn[16])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x08));
	
	return transmitPacket(context, ksn, 16, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_dukptEncryptECB(void *handler, const unsigned char iv[8], 
								const unsigned char data[], unsigned int dataLen,  unsigned char out[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x09));
	CHECK_RESULT(writePacketByte(context, 0x00));
	CHECK_RESULT(writePacketData(context, iv, 8));	
	CHECK_RESULT(writePacketData(context, data, dataLen));	
	
	return transmitPacket(context, out, dataLen, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_dukptEncryptCBC(void *handler, const unsigned char iv[8], 
								const unsigned char data[], unsigned int dataLen,  unsigned char out[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x32));	
	CHECK_RESULT(writePacketByte(context, 0x09));
	CHECK_RESULT(writePacketByte(context, 0x02));
	CHECK_RESULT(writePacketData(context, iv, 8));	
	CHECK_RESULT(writePacketData(context, data, dataLen));	
	
	return transmitPacket(context, out, dataLen, NULL);	
}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// SMART CARD COMMANDS
////////////////////////////////////////////////////////////////////////////////
PPSDK_ERR ppsdk_scInit(void *handler, unsigned char slot, unsigned char voltage, unsigned char protocol)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x33));	
	CHECK_RESULT(writePacketByte(context, 0x01));
	CHECK_RESULT(writePacketByte(context, slot));
	CHECK_RESULT(writePacketByte(context, voltage));
	CHECK_RESULT(writePacketByte(context, protocol));	
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_scCardPowerOn(void *handler, unsigned char slot, unsigned char atr[], unsigned int atrMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x33));	
	CHECK_RESULT(writePacketByte(context, 0x02));
	CHECK_RESULT(writePacketByte(context, slot));	
	
	return transmitPacket(context, atr, atrMaxLen, nbr);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_scCardPowerOff(void *handler, unsigned char slot)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x33));	
	CHECK_RESULT(writePacketByte(context, 0x03));
	CHECK_RESULT(writePacketByte(context, slot));	
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_scCheckCard(void *handler, unsigned char slot)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x33));	
	CHECK_RESULT(writePacketByte(context, 0x04));
	CHECK_RESULT(writePacketByte(context, slot));	
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_scCAPDU(void *handler, unsigned char slot, const unsigned char in[], unsigned char inLen, 
						unsigned char out[], unsigned char outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;
	
	CHECK_RESULT(constructPacket(context, 0x33));	
	CHECK_RESULT(writePacketByte(context, 0x05));
	CHECK_RESULT(writePacketByte(context, slot));	
	CHECK_RESULT(writePacketData(context, in, inLen));	
	
	return transmitPacket(context, out, outMaxLen, nbr);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_scShutdown(void *handler, unsigned char slot)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x33));	
	CHECK_RESULT(writePacketByte(context, 0x06));
	CHECK_RESULT(writePacketByte(context, slot));		
	
	return transmitPacket(context, NULL, 0, NULL);	
}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// MAGNETIC CARD COMMANDS
////////////////////////////////////////////////////////////////////////////////
PPSDK_ERR ppsdk_msStart(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x34));	
	CHECK_RESULT(writePacketByte(context, 0x01));	
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_msPull(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x34));	
	CHECK_RESULT(writePacketByte(context, 0x02));	
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_msGetCardData(void *handler, unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x34));	
	CHECK_RESULT(writePacketByte(context, 0x03));	
	CHECK_RESULT(writePacketByte(context, 0x87));	
	
	return transmitPacket(context, out, outMaxLen, nbr);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_msStop(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x34));	
	CHECK_RESULT(writePacketByte(context, 0x04));	
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_msGetEncryptedCardDataRSA(void *handler, unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x34));	
	CHECK_RESULT(writePacketByte(context, 0x07));		
	
	return transmitPacket(context, out, outMaxLen, nbr);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_msLoadRSAkey(void *handler, unsigned char key, const unsigned char data[], unsigned int dataLen)
{
	hcontext *context = (hcontext *)handler;

	if (data == NULL) return PPSDK_EINVPAR;

	CHECK_RESULT(constructPacket(context, 0x34));	
	CHECK_RESULT(writePacketByte(context, 0x08));		
	CHECK_RESULT(writePacketData(context, data, dataLen));		
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_msGetManualCardData(void *handler, unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x34));	
	CHECK_RESULT(writePacketByte(context, 0x09));		
	
	return transmitPacket(context, out, outMaxLen, nbr);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_msGetEncryptedCardDataCBC(void *handler, unsigned char key, unsigned int id,
										  unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x34));
	CHECK_RESULT(writePacketByte(context, 0x0A));
	CHECK_RESULT(writePacketByte(context, 0x87));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketByte(context, id << 24));
	CHECK_RESULT(writePacketByte(context, id << 16));
	CHECK_RESULT(writePacketByte(context, id << 8));
	CHECK_RESULT(writePacketByte(context, id));	
	
	return transmitPacket(context, out, outMaxLen, nbr);		
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_msGetMaskedCardData(void *handler, unsigned int unmaskedDigitsAtStart, 
									unsigned int unmaskedDigitsAtEnd, unsigned int unmaskedDigitsAfter,
									unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x34));
	CHECK_RESULT(writePacketByte(context, 0x0B));
	CHECK_RESULT(writePacketByte(context, unmaskedDigitsAtStart));	
	CHECK_RESULT(writePacketByte(context, unmaskedDigitsAtEnd));	
	CHECK_RESULT(writePacketByte(context, unmaskedDigitsAfter));	
	
	return transmitPacket(context, out, outMaxLen, nbr);	
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// DISPLAY / KEYBOARD COMMANDS
////////////////////////////////////////////////////////////////////////////////
PPSDK_ERR ppsdk_uiDrawString(void *handler, const char text[])
{
	hcontext *context = (hcontext *)handler;
	
	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x01));
	CHECK_RESULT(writePacketData(context, (unsigned char *)text, (unsigned int)strlen(text)));	
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiReadKey(void *handler, unsigned char *key)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x02));	
	
	return transmitPacket(context, key, 1, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiEnableNumericMode(void *handler, unsigned int x, unsigned int y, unsigned int lang, const char text[])
{
	hcontext *context = (hcontext *)handler;
	
	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x03));	
	CHECK_RESULT(writePacketByte(context, x));
	CHECK_RESULT(writePacketByte(context, y));
	CHECK_RESULT(writePacketByte(context, lang));
	CHECK_RESULT(writePacketData(context, (unsigned char *)text, (unsigned int)strlen(text)));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiDisableNumericMode(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x04));	
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiOpenTextWindow(void *handler, unsigned int x, unsigned int y, unsigned int ch, unsigned int cv, 
								 unsigned char font, unsigned char codetable)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x05));
	CHECK_RESULT(writePacketByte(context, x));
	CHECK_RESULT(writePacketByte(context, y));
	CHECK_RESULT(writePacketByte(context, ch));
	CHECK_RESULT(writePacketByte(context, cv));
	CHECK_RESULT(writePacketByte(context, font));
	CHECK_RESULT(writePacketByte(context, codetable));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiGetKeyboardControl(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x06));
	CHECK_RESULT(writePacketByte(context, 1));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiReleaseKeyboardControl(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x06));
	CHECK_RESULT(writePacketByte(context, 0));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiSetKeyTone(void *handler, unsigned int freq, unsigned int duration, unsigned char volume)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x07));
	CHECK_RESULT(writePacketByte(context, freq));
	CHECK_RESULT(writePacketByte(context, freq >> 8));
	CHECK_RESULT(writePacketByte(context, duration));
	CHECK_RESULT(writePacketByte(context, duration >> 8));
	CHECK_RESULT(writePacketByte(context, volume / 2));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiEnterPIN(void *handler, unsigned int mode, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x08));
	CHECK_RESULT(writePacketByte(context, mode));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiEnterPIN2(void *handler, unsigned int x, unsigned int y, char echo, unsigned int timeout, const char msg[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x09));
	CHECK_RESULT(writePacketByte(context, x));
	CHECK_RESULT(writePacketByte(context, y));
	CHECK_RESULT(writePacketByte(context, (unsigned char)echo));
	CHECK_RESULT(writePacketByte(context, timeout));
	CHECK_RESULT(writePacketData(context, (unsigned char *)msg, (unsigned int)strlen(msg)));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiVerifyPINOffline(void *handler, unsigned char type, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x0A));
	CHECK_RESULT(writePacketByte(context, type));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiVerifyPINOnline(void *handler, unsigned char mode, unsigned char pinfmt, unsigned char key, 
								  unsigned char variant[16], const char pan[], unsigned char pin[8])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x0A));
	CHECK_RESULT(writePacketByte(context, mode));
	CHECK_RESULT(writePacketByte(context, pinfmt));
	CHECK_RESULT(writePacketData(context, variant, 16));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketData(context, (unsigned char *)pan, (unsigned int)strlen(pan)));
	
	return transmitPacket(context, pin, 8, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiFillRectangle(void *handler, unsigned int x, unsigned int y, 
								unsigned int width, unsigned int height, unsigned int color)								  
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x0B));
	CHECK_RESULT(writePacketByte(context, x));
	CHECK_RESULT(writePacketByte(context, y));
	CHECK_RESULT(writePacketByte(context, width));
	CHECK_RESULT(writePacketByte(context, height));
	CHECK_RESULT(writePacketByte(context, color));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiSetContrast(void *handler, unsigned int contrast)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x0C));
	CHECK_RESULT(writePacketByte(context, contrast));

	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiDrawPixel(void *handler, unsigned int x, unsigned int y, unsigned int color)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x0D));
	CHECK_RESULT(writePacketByte(context, x));
	CHECK_RESULT(writePacketByte(context, y));
	CHECK_RESULT(writePacketByte(context, color));

	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiDrawBitmap(void *handler, unsigned int x, unsigned int y, 
								unsigned int width, unsigned int height, const unsigned char data[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x0E));
	CHECK_RESULT(writePacketByte(context, x));
	CHECK_RESULT(writePacketByte(context, y));
	CHECK_RESULT(writePacketByte(context, width));
	CHECK_RESULT(writePacketByte(context, height));
	CHECK_RESULT(writePacketData(context, data, (x * y) / 8));

	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiStartAnimation(void *handler, unsigned int x, unsigned int y, unsigned char anim, unsigned char mode)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x10));
	CHECK_RESULT(writePacketByte(context, 1));
	CHECK_RESULT(writePacketByte(context, anim));
	CHECK_RESULT(writePacketByte(context, x));
	CHECK_RESULT(writePacketByte(context, y));
	CHECK_RESULT(writePacketByte(context, mode));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiStopAnimation(void *handler, unsigned char anim)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x10));
	if (anim == 0xFF) {
		CHECK_RESULT(writePacketByte(context, 0x03));
	} else {
		CHECK_RESULT(writePacketByte(context, 0x00));
	}	
	CHECK_RESULT(writePacketByte(context, anim));
	CHECK_RESULT(writePacketByte(context, 0x00));
    CHECK_RESULT(writePacketByte(context, 0x00));
    CHECK_RESULT(writePacketByte(context, 0x00));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiSetButtonText(void *handler, unsigned char button, const char text[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x11));
	CHECK_RESULT(writePacketByte(context, button));
	CHECK_RESULT(writePacketData(context, (unsigned char *)text, (unsigned int)strlen(text)));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiInitScreen(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x23));	
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiEnterPINAsync(void *handler, unsigned int x, unsigned int y, char echo, unsigned int timeout, const char msg[])
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x20));
	CHECK_RESULT(writePacketByte(context, x));
	CHECK_RESULT(writePacketByte(context, y));
	CHECK_RESULT(writePacketByte(context, (unsigned char)echo));
	CHECK_RESULT(writePacketByte(context, timeout));
	CHECK_RESULT(writePacketData(context, (unsigned char *)msg, (unsigned int)strlen(msg)));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiCheckPIN(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x21));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_uiCancelPIN(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x35));
	CHECK_RESULT(writePacketByte(context, 0x22));
	
	return transmitPacket(context, NULL, 0, NULL);
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// SYSTEM COMMANDS
////////////////////////////////////////////////////////////////////////////////
PPSDK_ERR ppsdk_sysGetClock(void *handler, ppsdk_date *date)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x01));
	
	return transmitPacket(context, (unsigned char *)date, 6, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_sysSetDate(void *handler, const ppsdk_date *date)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x02));
	CHECK_RESULT(writePacketByte(context, date->year));
	CHECK_RESULT(writePacketByte(context, date->month));
	CHECK_RESULT(writePacketByte(context, date->day));
	CHECK_RESULT(writePacketByte(context, date->hour));
	CHECK_RESULT(writePacketByte(context, date->minute));
	CHECK_RESULT(writePacketByte(context, date->second));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_sysBeep(void *handler, unsigned int freq, unsigned int duration, unsigned char volume)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x03));
	CHECK_RESULT(writePacketByte(context, freq));
	CHECK_RESULT(writePacketByte(context, freq >> 8));
	CHECK_RESULT(writePacketByte(context, duration));
	CHECK_RESULT(writePacketByte(context, duration >> 8));
	CHECK_RESULT(writePacketByte(context, volume / 2));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_sysSetLED(void *handler, unsigned char led, unsigned int pattern)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x04));
	CHECK_RESULT(writePacketByte(context, led));
	CHECK_RESULT(writePacketByte(context, pattern));
	CHECK_RESULT(writePacketByte(context, pattern >> 8));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_sysShowStatus(void *handler, unsigned char value)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x05));
	CHECK_RESULT(writePacketByte(context, value));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_sysPlayMelody(void *handler, const unsigned char data[], unsigned int dataLen)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x06));
	CHECK_RESULT(writePacketData(context, data, dataLen));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_sysEraseFlash(void *handler, unsigned int address)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x07));
	CHECK_RESULT(writePacketByte(context, address));
	CHECK_RESULT(writePacketByte(context, address >> 8));
	CHECK_RESULT(writePacketByte(context, address >> 16));
	CHECK_RESULT(writePacketByte(context, address >> 24));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_sysWriteFlash(void *handler, unsigned int address, const unsigned char data[], unsigned int dataLen)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x08));
	CHECK_RESULT(writePacketByte(context, address));
	CHECK_RESULT(writePacketByte(context, address >> 8));
	CHECK_RESULT(writePacketByte(context, address >> 16));
	CHECK_RESULT(writePacketByte(context, address >> 24));
	CHECK_RESULT(writePacketByte(context, 0x00));
	CHECK_RESULT(writePacketData(context, data, dataLen));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_sysGetHardwareInfo(void *handler, ppsdk_hardinfo *hardInfo)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x0A));
	
	return transmitPacket(context, (unsigned char *)hardInfo, 3, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_sysTurnOff(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x0B));
	
	return transmitPacket(context, NULL, 0, NULL);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_sysEnableEvents(void *handler, unsigned int mask)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x36));
	CHECK_RESULT(writePacketByte(context, 0x10));
	CHECK_RESULT(writePacketByte(context, (unsigned char)mask));
    CHECK_RESULT(writePacketByte(context, (unsigned char)(mask >> 8)));
    CHECK_RESULT(writePacketByte(context, (unsigned char)(mask >> 16)));
    CHECK_RESULT(writePacketByte(context, (unsigned char)(mask >> 24)));
	
	return transmitPacket(context, NULL, 0, NULL);
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// EMV2 COMMANDS (NECOM+ KERNEL)
////////////////////////////////////////////////////////////////////////////////
PPSDK_ERR ppsdk_emvInit(void *handler, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x01));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvDeinitialize(void *handler, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x02));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvGetVersion(void *handler, ppsdk_emv_ver *emvver, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x03));
	
	return emvTransmitPacket(context, (unsigned char *)emvver, 15, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvValidateATR(void *handler, unsigned char type, const unsigned char atr[] , unsigned int atrLen, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x04));
	CHECK_RESULT(writePacketByte(context, type));
	CHECK_RESULT(writePacketData(context, atr, atrLen));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvLoadAppList(void *handler, unsigned char method, const char incblk, const ppsdk_emv_app appList[], unsigned int count, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;
	unsigned int n = 0;
	
	if (appList == NULL) return PPSDK_EINVPAR;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x05));
	CHECK_RESULT(writePacketByte(context, method));
	CHECK_RESULT(writePacketByte(context, incblk));
	CHECK_RESULT(writePacketByte(context, count));

	while (n < count) {
		CHECK_RESULT(writePacketByte(context, appList[n].aidLen));
		CHECK_RESULT(writePacketData(context, appList[n].aid, appList[n].aidLen));
		CHECK_RESULT(writePacketByte(context, appList[n].match));		
		CHECK_RESULT(writePacketData(context, (unsigned char *)appList[n].label, (unsigned int)(strlen(appList[n].label) + 1)));
		n++;
	}

	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvGetCommonAppList(void *handler, unsigned char *confirm, ppsdk_emv_app app[], unsigned int maxApp, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;
	unsigned char *p = NULL;
	unsigned char c = 0;
	unsigned char n = 0;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x06));

	CHECK_RESULT(emvTransmitPacket(context, NULL, 0, NULL, sw));
	
	p = &context->packet[PACKET_HEADER_SIZE + 2];
	c = *p++;
	n = *p++;	
	
	if (confirm != NULL) *confirm = c;

	if (app != NULL) {
		if (n > maxApp) return PPSDK_EINVLEN;

		memset(app, 0, sizeof(ppsdk_emv_app) * maxApp);

		while (n > 0) {
			unsigned char nameLen = 0;

			app->aidLen = *p++;
			memcpy(app->aid, p, app->aidLen);
			p+= app->aidLen;
			nameLen = *p++;
			memcpy(app->label, p, nameLen);
			p+=nameLen;
			app++;
			n--;
		}
	}

	return PPSDK_ENOERR;	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvInitialAppProcessing(void *handler, const unsigned char aid[] , unsigned int aidLen, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x07));
	CHECK_RESULT(writePacketData(context, aid, aidLen));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvReadAppData(void *handler, const unsigned short tagList[] , unsigned int count, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;
	unsigned int n = 0;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x08));
	
	while (n < count) {
		CHECK_RESULT(writePacketByte(context, (unsigned char)tagList[n]));		
		CHECK_RESULT(writePacketByte(context, (unsigned char)(tagList[n] >> 8)));
		n++;
	}

	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvAutentication(void *handler, unsigned char check, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x09));
	CHECK_RESULT(writePacketByte(context, check));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvProcessRestrictions(void *handler, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x0A));	
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvTerminalRisk(void *handler, unsigned char force, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x0B));	
	CHECK_RESULT(writePacketByte(context, force));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvGetAuthMethod(void *handler, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x0C));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvSetAuthResult(void *handler, unsigned char authRes, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x0D));
	CHECK_RESULT(writePacketByte(context, authRes));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv_verifyPinOFF(void *handler, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x0E));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvGenerateCertificate(void *handler, unsigned char type, unsigned char risk, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x0F));
	CHECK_RESULT(writePacketByte(context, type));
	CHECK_RESULT(writePacketByte(context, risk));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvMakeTransactionDesision(void *handler, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x10));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvMakeDefaultDecision(void *handler, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x11));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvAuthenticateIssuer(void *handler, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x12));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvScriptProcessing(void *handler, unsigned char script, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x13));
	CHECK_RESULT(writePacketByte(context, script));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvUpdateTVR(void *handler, unsigned char byteIdx, unsigned char bitIdx, unsigned char value, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x14));
	CHECK_RESULT(writePacketByte(context, byteIdx));
	CHECK_RESULT(writePacketByte(context, bitIdx));
	CHECK_RESULT(writePacketByte(context, value));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvUpdateTSI(void *handler, unsigned char byteIdx, unsigned char bitIdx, unsigned char value, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x15));
	CHECK_RESULT(writePacketByte(context, byteIdx));
	CHECK_RESULT(writePacketByte(context, bitIdx));
	CHECK_RESULT(writePacketByte(context, value));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvCheckTVR(void *handler, unsigned char byteIdx, unsigned char bitIdx, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x16));
	CHECK_RESULT(writePacketByte(context, byteIdx));
	CHECK_RESULT(writePacketByte(context, bitIdx));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvCheckTSI(void *handler, unsigned char byteIdx, unsigned char bitIdx, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x17));
	CHECK_RESULT(writePacketByte(context, byteIdx));
	CHECK_RESULT(writePacketByte(context, bitIdx));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvLoadPublicKey(void *handler, const unsigned char rid[5], unsigned char idx, 
								 const unsigned char mod[], unsigned int modLen, 
								 const unsigned char exp[], unsigned int expLen, 
								 const unsigned char hash[20], unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x18));
	CHECK_RESULT(writePacketData(context, rid, 5));
	CHECK_RESULT(writePacketByte(context, idx));
	CHECK_RESULT(writePacketByte(context, modLen));
	CHECK_RESULT(writePacketData(context, mod, modLen));
	CHECK_RESULT(writePacketByte(context, expLen));
	CHECK_RESULT(writePacketData(context, exp, expLen));
	CHECK_RESULT(writePacketData(context, hash, 20));

	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvRemovePublicKey(void *handler, const unsigned char rid[5], unsigned char idx, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x19));
	CHECK_RESULT(writePacketData(context, rid, 5));
	CHECK_RESULT(writePacketByte(context, idx));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvSetDataAsBinary(void *handler, unsigned short tag, const unsigned char data[], unsigned int dataLen, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x1A));
	CHECK_RESULT(writePacketByte(context, (unsigned char)tag));
	CHECK_RESULT(writePacketByte(context, (unsigned char)(tag >> 8)));
	CHECK_RESULT(writePacketData(context, data, dataLen));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvSetDataAsString(void *handler, unsigned short tag, const char data[], unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x1B));
	CHECK_RESULT(writePacketByte(context, (unsigned char)tag));
	CHECK_RESULT(writePacketByte(context, (unsigned char)(tag >> 8)));	
	CHECK_RESULT(writePacketData(context, data, (unsigned int)strlen(data) + 1));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvGetDataAsBinary(void *handler, unsigned short tag, unsigned char out[], unsigned int outMaxLen, unsigned int *nbr, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x1C));
	CHECK_RESULT(writePacketByte(context, (unsigned char)tag));
	CHECK_RESULT(writePacketByte(context, (unsigned char)(tag >> 8)));
	
	return emvTransmitPacket(context, out, outMaxLen, nbr, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvGetDataAsString(void *handler, unsigned short tag, char out[], unsigned int outMaxLen, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x1D));
	CHECK_RESULT(writePacketByte(context, (unsigned char)tag));
	CHECK_RESULT(writePacketByte(context, (unsigned char)(tag >> 8)));
	
	if (out != NULL) {
		memset(out, 0, outMaxLen); 
	}

	return emvTransmitPacket(context, (unsigned char *)out, outMaxLen, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvGetDataDetails(void *handler, unsigned short tag, unsigned char *type, unsigned int *maxLen, unsigned int *curLen, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x1E));
	CHECK_RESULT(writePacketByte(context, (unsigned char)tag));
	CHECK_RESULT(writePacketByte(context, (unsigned char)(tag >> 8)));
	
	CHECK_RESULT(emvTransmitPacket(context, NULL, 0, NULL, sw));
	
	if (sw == PPSDK_EMV_STATUS_SUCCESS) {
		if (type != NULL) *type = context->packet[PACKET_HEADER_SIZE + 2];
		if (maxLen != NULL) *maxLen = byteArrayToInt(&context->packet[PACKET_HEADER_SIZE + 2 + 2], 2);
		if (curLen != NULL) *curLen = byteArrayToInt(&context->packet[PACKET_HEADER_SIZE + 2 + 4], 2);
	}

	return PPSDK_ENOERR;
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvSetBypassMode(void *handler, unsigned char mode, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x1F));
	CHECK_RESULT(writePacketByte(context, mode));	
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvSetTags(void *handler, const unsigned char tlv[], unsigned int tlvLen, unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x80));
	CHECK_RESULT(writePacketData(context, tlv, tlvLen));
	
	return emvTransmitPacket(context, NULL, 0, NULL, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvGetTags(void *handler, const unsigned char tag[], unsigned int tagLen, 
						   unsigned char tlv[], unsigned int tlvMaxLen, unsigned int *nbr,
						   unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x81));
	CHECK_RESULT(writePacketData(context, tag, tagLen));
	
	return emvTransmitPacket(context, tlv, tlvMaxLen, nbr, sw);
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_emvGetTagsEncrypted(void *handler, unsigned char key, unsigned int id, const unsigned char tag[], unsigned int tagLen, 
						   unsigned char data[], unsigned int dataMaxLen, unsigned int *nbr,
						   unsigned short *sw)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x37));
	CHECK_RESULT(writePacketByte(context, 0x81));
	CHECK_RESULT(writePacketByte(context, key));
	CHECK_RESULT(writePacketByte(context, (unsigned char)id));
	CHECK_RESULT(writePacketByte(context, (unsigned char)(id >> 8)));
	CHECK_RESULT(writePacketByte(context, (unsigned char)(id >> 16)));
	CHECK_RESULT(writePacketByte(context, (unsigned char)(id >> 24)));	
	CHECK_RESULT(writePacketData(context, tag, tagLen));
	
	return emvTransmitPacket(context, data, dataMaxLen, nbr, sw);
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// BARCODE COMMANDS
////////////////////////////////////////////////////////////////////////////////
PPSDK_ERR ppsdk_bcPowerOn(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x38));	
	CHECK_RESULT(writePacketByte(context, 0x00));
	CHECK_RESULT(writePacketByte(context, 0x01));		
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_bcPowerOff(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x38));	
	CHECK_RESULT(writePacketByte(context, 0x00));
	CHECK_RESULT(writePacketByte(context, 0x00));		
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_bcPull(void *handler)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x38));	
	CHECK_RESULT(writePacketByte(context, 0x02));		
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_bcGetData(void *handler, unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x38));	
	CHECK_RESULT(writePacketByte(context, 0x03));		
	
	return transmitPacket(context, out, outMaxLen, nbr);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_bcStartRead(void *handler, unsigned char mode)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x38));	
	CHECK_RESULT(writePacketByte(context, 0x05));		
	CHECK_RESULT(writePacketByte(context, mode));
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_bcSetInitCommand(void *handler, const unsigned char data[], unsigned int dataLen)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x38));	
	CHECK_RESULT(writePacketByte(context, 0x06));		
	CHECK_RESULT(writePacketData(context, data, dataLen));		
	
	return transmitPacket(context, NULL, 0, NULL);	
}
//-----------------------------------------------------------------------------
PPSDK_ERR ppsdk_bcSetSound(void *handler, unsigned int freq, unsigned int duration, unsigned char volume)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x38));
	CHECK_RESULT(writePacketByte(context, 0x07));
	CHECK_RESULT(writePacketByte(context, freq));
	CHECK_RESULT(writePacketByte(context, freq >> 8));
	CHECK_RESULT(writePacketByte(context, duration));
	CHECK_RESULT(writePacketByte(context, duration >> 8));
	CHECK_RESULT(writePacketByte(context, volume / 2));
	
	return transmitPacket(context, NULL, 0, NULL);
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// EMV 2
////////////////////////////////////////////////////////////////////////////////    
// Helper transmit method
static PPSDK_ERR emv2Transmit(void *handler, int command, 
							 const unsigned char prev[], unsigned int prevLen, 
							 const unsigned char data[], unsigned int dataLen, 
							 unsigned char out[], unsigned int outMaxLen,
							 unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	CHECK_RESULT(constructPacket(context, 0x3E));	
	CHECK_RESULT(writePacketByte(context, command));
	CHECK_RESULT(writePacketByte(context, 0));
	CHECK_RESULT(writePacketData(context, prev, prevLen));			
	CHECK_RESULT(writePacketData(context, data, dataLen));			
	CHECK_RESULT(transmitPacket(context, out, outMaxLen, nbr));	
	
	if (nbr != NULL) {
		*nbr-=2;

		if (out != NULL) {
			memmove(out, out + 2, *nbr);
		}		
	}

	return PPSDK_ENOERR;   
}
//-----------------------------------------------------------------------------    
PPSDK_ERR ppsdk_emv2LoadConfiguration(void *handler, int type, const unsigned char data[], unsigned int dataLen)
{
	hcontext *context = (hcontext *)handler;	
	unsigned char buffer[] = {	
		(unsigned char)(type),
		(unsigned char)(dataLen >> 24),
		(unsigned char)(dataLen >> 16),
		(unsigned char)(dataLen >> 8),
		(unsigned char)(dataLen),
	};
	CHECK_RESULT(emv2Transmit(handler, 0x51, NULL, 0, buffer, sizeof(buffer), NULL, 0, NULL));

	// Send data
	while (dataLen > 0) {
		int size = dataLen;		
		if (size > 256) {
			size = 256;
		}

		CHECK_RESULT(constructPacket(context, 0x52));
		CHECK_RESULT(writePacketData(context, data, size));
		CHECK_RESULT(emv2Transmit(handler, 0x52, NULL, 0, data, size, NULL, 0, NULL));
		data +=size;
		dataLen -= size;					
	}

    return emv2Transmit(handler, 0x53, NULL, 0, NULL, 0, NULL, 0, NULL);        
}    
// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2LoadContactlessConfiguration(void *handler,const unsigned char data[], unsigned int dataLen)
{
	hcontext *context = (hcontext *)handler;
    return ppsdk_emv2LoadConfiguration(handler, PPSDK_EMV2_CONFIG_CONTACTLESS, data, dataLen);
}
// ---------------------------------------------------------------------------------------------	
PPSDK_ERR ppsdk_emv2LoadContactConfiguration(void *handler,const unsigned char data[], unsigned int dataLen)
{
	hcontext *context = (hcontext *)handler;
    return ppsdk_emv2LoadConfiguration(handler, PPSDK_EMV2_CONFIG_CHIP, data, dataLen);
}    
// ---------------------------------------------------------------------------------------------    
PPSDK_ERR ppsdk_emv2LoadContactlessCAPK(void *handler,const unsigned char data[], unsigned int dataLen)
{
	hcontext *context = (hcontext *)handler;
    return ppsdk_emv2LoadConfiguration(handler, PPSDK_EMV2_CONFIG_CL_CA_KEYS, data, dataLen);
}
// ---------------------------------------------------------------------------------------------    
PPSDK_ERR ppsdk_emv2LoadContactCAPK(void *handler,const unsigned char data[], unsigned int dataLen)
{
	hcontext *context = (hcontext *)handler;
    return ppsdk_emv2LoadConfiguration(handler, PPSDK_EMV2_CONFIG_CHIP_CA_KEYS, data, dataLen);
}       
// ---------------------------------------------------------------------------------------------
static PPSDK_ERR parseEMV2Info(const unsigned char data[], unsigned int dataLen, PPSDK_EMV2Info* info)
{	
	bertlv_u8 *ptr;
	bertlv_u32 len;

	if (info != NULL) {		
		memset(info, 0, sizeof(PPSDK_EMV2Info));		

		len = bertlv_find(data, dataLen, 0xC1, &ptr);
		if (len > 0) {
			bertlv_decode(ptr, len, NULL, &len, &ptr);
			info->configurationVersion = byteArrayToIntBE(ptr, len);
            info->contactlessConfigurationVersion = info->configurationVersion;				
		} else {
			return PPSDK_EGENERAL;
		}

		len = bertlv_find(data, dataLen, 0xC2, &ptr);
		if (len > 0) {
			bertlv_decode(ptr, len, NULL, &len, &ptr);
			info->contactConfigurationVersion = byteArrayToIntBE(ptr, len);
		}
        
		len = bertlv_find(data, dataLen, 0xC3, &ptr);
		if (len > 0) {
			bertlv_decode(ptr, len, NULL, &len, &ptr);
			info->contactlessCAPKVersion = byteArrayToIntBE(ptr, len);
		}

        len = bertlv_find(data, dataLen, 0xC4, &ptr);
		if (len > 0) {
			bertlv_decode(ptr, len, NULL, &len, &ptr);
			info->contactCAPKVersion = byteArrayToIntBE(ptr, len);
		}
       
		len = bertlv_find(data, dataLen, 0xC5, &ptr);
		if (len > 0) {
			bertlv_decode(ptr, len, NULL, &len, &ptr);
			info->emvKernelVersion = byteArrayToIntBE(ptr, len);
		}       
	}

	return PPSDK_ENOERR;
}

// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2GetInfo(void *handler, PPSDK_EMV2Info* info) 
{
	hcontext *context = (hcontext *)handler;
	unsigned char buffer[32];
	unsigned int nbr;

    CHECK_RESULT(emv2Transmit(context, 0x02, NULL, 0, NULL, 0, buffer, sizeof(buffer), &nbr));

	if (info != NULL) {
		CHECK_RESULT(parseEMV2Info(buffer, nbr, info));
	}
	
    return PPSDK_ENOERR;
}    
// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2StartTransaction(void *handler, int interfaces, int flags, const unsigned char data[], 
									 unsigned int dataLen, int timeout)
{
	hcontext *context = (hcontext *)handler;
	unsigned char buffer[2];
	buffer[0] = (unsigned char)interfaces;
	buffer[1] = (unsigned char)timeout;
    return emv2Transmit(context, 0x03, buffer, sizeof(buffer), data, dataLen, NULL, 0, NULL);
}
// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2SetOnlineResult(void *handler, const unsigned char tags[], unsigned int tagsLen)
{
	hcontext *context = (hcontext *)handler;
    return emv2Transmit(context, 0x04, NULL, 0, tags, tagsLen, NULL, 0, NULL);
}
// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2CancelTransaction(void *handler) 
{
	hcontext *context = (hcontext *)handler;
    return emv2Transmit(context, 0x05, NULL, 0, NULL, 0, NULL, 0, NULL);
}
// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2Initialize(void *handler)
{
	hcontext *context = (hcontext *)handler;
    return emv2Transmit(context, 0x61, NULL, 0, NULL, 0, NULL, 0, NULL);        
}
// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2Deinitialise(void *handler)
{
	hcontext *context = (hcontext *)handler;
    return emv2Transmit(context, 0x62, NULL, 0, NULL, 0, NULL, 0, NULL);        
}
// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2GetTrackData(void *handler, int format, int keyID, unsigned char out[], 
								 unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	unsigned char buffer[] = {
		(unsigned char)(0xC1),
		(unsigned char)(4),
		(unsigned char)(format >> 24),
		(unsigned char)(format >> 16),
		(unsigned char)(format >> 8),
		(unsigned char)(format),

		(unsigned char)(0xC2),
		(unsigned char)(4),
		(unsigned char)(keyID >> 24),
		(unsigned char)(keyID >> 16),
		(unsigned char)(keyID >> 8),
		(unsigned char)(keyID),
	};

    return emv2Transmit(context, 0x06, NULL, 0, buffer, sizeof(buffer), out, outMaxLen, nbr);
}

// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2GetTrackData2(void *handler, int format, int keyID, int showExpiration, int
        unmaskedDigitsAtStart, int unmaskedDigitsAtEnd, 
		unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	unsigned char buffer[] = {
		(unsigned char)(0xC1), 
		(unsigned char)(4),
		(unsigned char)(format >> 24),
		(unsigned char)(format >> 16),
		(unsigned char)(format >> 8),
		(unsigned char)(format),

		(unsigned char)(0xC2),
		(unsigned char)(4),
		(unsigned char)(keyID >> 24),
		(unsigned char)(keyID >> 16),
		(unsigned char)(keyID >> 8),
		(unsigned char)(keyID),

		(unsigned char)(0xC3),
		(unsigned char)(1),
		(unsigned char)(showExpiration ? 1 : 0),
		
		(unsigned char)(0xC4),
		(unsigned char)(1),
		(unsigned char)(unmaskedDigitsAtStart),

		(unsigned char)(0xC5),
		(unsigned char)(1),
		(unsigned char)(unmaskedDigitsAtEnd),
	};

	return emv2Transmit(context, 0x06, NULL, 0, buffer, sizeof(buffer), out, outMaxLen, nbr);
}

// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2GetTrackData3(void *handler, int format, int keyID, int sessionID,
							   unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	unsigned char buffer[] = {
		(unsigned char)(0xC1), 
		(unsigned char)(4),
		(unsigned char)(format >> 24),
		(unsigned char)(format >> 16),
		(unsigned char)(format >> 8),
		(unsigned char)(format),

		(unsigned char)(0xC2),
		(unsigned char)(4),
		(unsigned char)(keyID >> 24),
		(unsigned char)(keyID >> 16),
		(unsigned char)(keyID >> 8),
		(unsigned char)(keyID),

		(unsigned char)(0xC3),
		(unsigned char)(4),
		(unsigned char)(sessionID >> 24),
		(unsigned char)(sessionID >> 16),
		(unsigned char)(sessionID >> 8),
		(unsigned char)(sessionID),
	};

	return emv2Transmit(context, 0x06, NULL, 0, buffer, sizeof(buffer), out, outMaxLen, nbr);    
}

// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2GetTagsEncrypted(void *handler, const unsigned char tags[], unsigned int tagsLen, 
								   int format, int keyType, int keyIndex, int packetID,
								   unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{

	hcontext *context = (hcontext *)handler;

	int keyID = (keyType  << 10) | keyIndex;

	unsigned char buffer[] = {
		(unsigned char)(0xC1), 
		(unsigned char)(4),
		(unsigned char)(format >> 24),
		(unsigned char)(format >> 16),
		(unsigned char)(format >> 8),
		(unsigned char)(format),

		(unsigned char)(0xC2),
		(unsigned char)(4),
		(unsigned char)(keyID >> 24),
		(unsigned char)(keyID >> 16),
		(unsigned char)(keyID >> 8),
		(unsigned char)(keyID),

		(unsigned char)(0xC4),
		(unsigned char)(4),
		(unsigned char)(packetID >> 24),
		(unsigned char)(packetID >> 16),
		(unsigned char)(packetID >> 8),
		(unsigned char)(packetID),	

		(unsigned char)(0xC3),
		(unsigned char)(0x82),		
		(unsigned char)(tagsLen >> 8),
		(unsigned char)(tagsLen),	
		// Can't put tag data here
	};

	return emv2Transmit(context, 0x07, buffer, sizeof(buffer), tags, tagsLen, out, outMaxLen, nbr);    
}

// ---------------------------------------------------------------------------------------------
PPSDK_ERR ppsdk_emv2GetTagsPlain(void *handler, const unsigned char tags[], unsigned int tagsLen,
								 unsigned char out[], unsigned int outMaxLen, unsigned int *nbr)
{
	hcontext *context = (hcontext *)handler;

	unsigned char buffer[] = {
		(unsigned char)(0xC3), 
		(unsigned char)(0x82),
		(unsigned char)(tagsLen >> 8),
		(unsigned char)(tagsLen),
		// Can't put tag data here
	};

    return emv2Transmit(context, 0x07, buffer, sizeof(buffer), tags, tagsLen, out, outMaxLen, nbr);    
}
// ---------------------------------------------------------------------------------------------